@@ -17,6 +17,8 @@ import (
 	"path/filepath"
 	"strings"
 	"syscall"
+
+	"launchpad.net/goget-ubuntu-touch/sysutils"
 )
 
 // This program is free software: you can redistribute it and/or modify it
@@ -39,6 +41,16 @@ func init() {
 	}
 }
 
+const (
+	hardwareFileName = "hardware.yaml"
+	kernelFileName   = "vmlinuz"
+	initrdFileName   = "initrd.img"
+)
+
+const (
+	partLayoutSystemAB = "system-AB"
+)
+
 var (
 	syscallSync = syscall.Sync
 )
@@ -62,8 +74,8 @@ type SystemImage interface {
 type CoreImage interface {
 	Image
 	SystemImage
-	SetupBoot(oemRootPath string) error
-	FlashExtra(oemRootPath, devicePart string) error
+	SetupBoot() error
+	FlashExtra() error
 }
 
 type HardwareDescription struct {
@@ -112,10 +124,28 @@ type OemDescription struct {
 	} `yaml:"oem,omitempty"`
 
 	Config map[string]interface{} `yaml:"config,omitempty"`
+
+	rootDir string
 }
 
-func (o OemDescription) InstallPath(rootPath string) (string, error) {
-	glob, err := filepath.Glob(fmt.Sprintf("%s/oem/%s/%s", rootPath, o.Name, o.Version))
+func (o *OemDescription) SetRoot(rootDir string) {
+	o.rootDir = rootDir
+}
+
+// SystemParts returns the system labels depending on the partition layout.
+//
+// The default is to return a flat structure for any unknown layout.
+func (o *OemDescription) SystemParts() []string {
+	switch o.OEM.Hardware.PartitionLayout {
+	case partLayoutSystemAB:
+		return []string{"a", "b"}
+	default:
+		return []string{""}
+	}
+}
+
+func (o OemDescription) InstallPath() (string, error) {
+	glob, err := filepath.Glob(fmt.Sprintf("%s/oem/%s/%s", o.rootDir, o.Name, o.Version))
 	if err != nil {
 		return "", err
 	}
@@ -165,6 +195,7 @@ type BaseImage struct {
 	parts     []partition
 	partCount int
 	size      int64
+	rootSize  int
 }
 
 // Mount mounts the image. This also maps the loop device.
@@ -399,6 +430,56 @@ func (img BaseImage) BaseMount() string {
 	return img.baseMount
 }
 
+func (img *BaseImage) GenericBootSetup(bootPath string) error {
+	// origins
+	hardwareYamlPath := filepath.Join(img.baseMount, hardwareFileName)
+	kernelPath := filepath.Join(img.baseMount, img.hardware.Kernel)
+	initrdPath := filepath.Join(img.baseMount, img.hardware.Initrd)
+
+	// populate both A/B
+	for _, part := range img.oem.SystemParts() {
+		path := filepath.Join(bootPath, part)
+
+		printOut("Setting up", path)
+
+		if err := os.MkdirAll(path, 0755); err != nil {
+			return err
+		}
+
+		if err := sysutils.CopyFile(hardwareYamlPath, filepath.Join(path, hardwareFileName)); err != nil {
+			return err
+		}
+
+		if err := sysutils.CopyFile(kernelPath, filepath.Join(path, kernelFileName)); err != nil {
+			return err
+		}
+
+		if err := sysutils.CopyFile(initrdPath, filepath.Join(path, initrdFileName)); err != nil {
+			return err
+		}
+	}
+
+	oemRoot, err := img.oem.InstallPath()
+	if err != nil {
+		return err
+	}
+
+	return setupBootAssetFiles(bootPath, oemRoot, img.oem.OEM.Hardware.BootAssets.Files)
+}
+
+func (img *BaseImage) FlashExtra() error {
+	oemRoot, err := img.oem.InstallPath()
+	if err != nil {
+		return err
+	}
+
+	if bootAssets := img.oem.OEM.Hardware.BootAssets; bootAssets != nil {
+		return setupBootAssetRawFiles(img.location, oemRoot, bootAssets.RawFiles)
+	}
+
+	return nil
+}
+
 func printOut(args ...interface{}) {
 	if debugPrint {
 		fmt.Println(args...)
